class FileExistsError(Exception):
    def __init__(self, element, message="File doesn's exist"):
        self.element = element
        self.message = message

        super().__init__(self.message)
    def __str__(self):
        return f'"{self.element}": {self.message}'

class UnexpectedType(Exception):
    def __init__(self, element, message="Unexpected Type"):
        self.element = element
        self.message = message

        super().__init__(self.message)
    def __str__(self):
        return f'"{self.element}": {self.message}'

class ArgumentTypeError(Exception):
    def __init__(self, element, message="Unexpected Type"):
        self.element = element
        self.message = message

        super().__init__(self.message)
    def __str__(self):
        return f'"{self.element}": {self.message}'

class NotPyonFile(Exception):
    def __init__(self, element, message="is not pyon file"):
        self.element = element
        self.message = message

        super().__init__(self.message)
    def __str__(self):
        return f'"{self.element}": {self.message}'


import os
from typing import Any

from .errors import *
from .decoder import PYONDecoder
from .encoder import PYONEncoder
from .dotnotation import DotNotation
from .builtins import is_pyon, is_json, convert, convert_json_to_pyon

def dumps(obj):
    '''
    convert `obj` to string
    '''
    return PYONEncoder(obj).encode()

def loads(string:str):
    '''
    convert `string` (pyon) to python dict
    '''
    return PYONDecoder(string).decode()

class Read:
    def __init__(self, filepath:str, encoding:str='utf-8'):
        if not os.path.isfile(filepath):
            raise FileExistsError(filepath)
        if not filepath.endswith('.pyon') and not filepath.endswith('.ndb'):
            raise NotPyonFile(filepath)
        
        self.__filepath = filepath
        self.encoding = encoding

        with open(self.__filepath, 'r', encoding=self.encoding) as file:
            self.__decoder = PYONDecoder(file.read())
            self.__file_data = self.__decoder.decode()

    def write(self, obj):
        '''
        write `obj` to file

        `obj` can be pyon string or json string or python dict

        return `True` on success
        '''
        filepath = self.__filepath
        obj_as_string = dumps(obj)
        
        # types handling
        if isinstance(obj, dict) or is_pyon(obj_as_string):
            with open(filepath, 'w', encoding=self.encoding) as file:
                file.write(obj_as_string)

            return True
        if is_json(obj_as_string):
            with open(filepath, 'w', encoding=self.encoding) as file:
                converted = str(convert(obj_as_string))
                file.write(converted)

            return True
        
        with open(filepath, 'w', encoding=self.encoding) as file:
            file.write(obj_as_string)

    @property
    def readfile(self) -> Any:
        '''
        reads file while still updating it (if a change was occurred)
        '''
        with open(self.__filepath, 'r', encoding=self.encoding) as file:
            self.__decoder = PYONDecoder(file.read())
            self.__file_data = self.__decoder.decode()

        return self.__file_data

    @property
    def filepath(self):
        '''
        return filepath
        '''
        return os.path.abspath(self.__filepath)

    # Magic methods
    def __repr__(self):
        args = []
        args.append(f'filepath={self.filepath}')

        return f'{self.__class__.__qualname__}({", ".join(args)})'

    def __add__(self, other):
        '''
        Return self+value.
        '''
        self_data = self.readfile
        try:
            other_data = other.readfile
        except AttributeError:
            other_data = other


        if isinstance(self_data, dict) and isinstance(other_data, dict):
            return {**self_data, **other_data}
        
        if isinstance(self_data, (list, int, str)) and isinstance(other_data, (list, int, str)):
            return self_data + other_data

    def __eq__(self, other) -> bool:
        self_data = self.readfile
        try:
            other_data = other.readfile
        except AttributeError:
            other_data = other

        return self_data == other_data

    def __ne__(self, other) -> bool:
        self_data = self.readfile
        try:
            other_data = other.readfile
        except AttributeError:
            other_data = other

        return self_data != other_data


import types

class DotNotation(types.SimpleNamespace):
    def __init__(self, dictionary, **kwargs):
        super().__init__(**kwargs)
        self.__dictionary = dictionary

        for key, value in dictionary.items():
            if isinstance(value, dict):
                setattr(self, key, DotNotation(value))
            else:
                self.__setattr__(key, value)

    def to_dict(self) -> dict:
        return self.__dictionary

    def __repr__(self) -> str:
        return f'{self.__dictionary}'


from requests import get
from typing import Any
from warnings import warn
import os.path
from json import loads, dumps

from .errors import *

def fetch(url_or_path:str) -> Any:
   if os.path.isfile(url_or_path): # if "url_or_path" is a file
      with open(url_or_path, 'r', encoding='utf-8') as file:
         file_data = file.read()

         if is_pyon(file_data) or is_json(file_data):
            return convert(file_data)
            
         warn('Warning: The path given is not either pyon or json, it will import the file text for you')
         return file.read()

   # if it's a url (it could not be a url but requests will raise an Exception)
   url_content = get(url_or_path).content.decode('utf-8')
   if is_pyon(url_content) or is_json(url_content):
      return convert(url_content)
   
   warn('Warning: The URL given is not either pyon or json, it will import the html string for you')
   return url_content

builtins = {'fetch': fetch}
def is_json(obj_as_string:str):
    '''
    checks if `obj_as_string` is json
    return `bool`
    '''
    try:
        loads(obj_as_string)
    except ValueError:
        return False
    except TypeError:
        return False
    
    return True

def is_pyon(obj_as_string:str):
    '''
    checks if `obj_as_string` is pyon
    return `bool`
    '''
    try:
        eval(obj_as_string, builtins)
    except ValueError:
        return False
    except SyntaxError:
        return False

    return True

def convert(string:str):
    """
    convert `string` to a `dict`
    works on json

    ```python
    pyon = '''
    {
        "user1": {
            "username": "nawaf",
            "email": "nawaf@domain.com",
            "verified": True
        }
    }
    '''

    converted = pyonr.convert(pyon)

    type(converted) # <class 'dict'>
    ```
    """
    if not string:
        raise ArgumentTypeError(string, 'Was not expecting empty string')

    if is_json(string):
        return loads(string)
    if is_pyon(string):
        return eval(string, builtins)

    return None

def convert_json_to_pyon(string):
    '''
    convert json string to pyon string
    '''
    obj_as_dict = loads(string)
    obj_as_pyon = convert(str(obj_as_dict))

    return obj_as_pyon


__all__ = ['PYONEncoder']

def _convert(obj):
    return str(obj)

class PYONEncoder(object):
    def __init__(self, obj):
        self.obj = obj

    def encode(self):
        return _convert(self.obj)


import ast
import json

from .errors import *
from .builtins import *

class PYONDecoder:
    def __init__(self, obj_as_string:str):
        if not isinstance(obj_as_string, str):
            raise UnexpectedType(obj_as_string, f'Expected `str` not ({type(obj_as_string)})')

        self.obj = obj_as_string

    def decode(self):
        obj = self.obj
        if not obj:
            return None

        return convert(obj)
    
    def decode_json(self):
        '''
        convert obj to a json obj (string)
        '''
        obj = self.obj
        if not obj:
            return '{}'

        return json.dumps(convert(obj))